"""
Dispersions app

App for dispersion analysis

"""
from analyses.analysisbase import models as base_models
from analyses.main import analysesmodules

from django.db import models
from django.db.models import signals as django_signals
from django.utils.functional import curry
from django.utils.translation import ugettext_lazy as _

from installations import models as installations_models

from utils.exchange import exchange
from utils.search.models import searches, Search
from utils.strings import add_suffix_to_string
from utils.trees import models as trees_models
from utils.versioning import models as versioning_models
from utils.versioning import versionedmptt

class AppConfig(base_models.AppConfig):
    namespace = "dispersions"
    tree_root_elementtype = "dispersion"
    analysistype =  _('Dispersion control')
    scenario_displaypermission = 'dispersions.display_dispersionscenarios'
    scenario_editpermission = 'dispersions.edit_dispersionscenarios'
    suggestionlist_displaypermission = 'dispersions.display_dispersionsuggestionlists'
    suggestionlist_editpermission = 'dispersions.edit_dispersionsuggestionlists'
    listurl = 'dispersions_scenarios'
    addurl = 'dispersions_scenario_add'
    consequence_label = _('Consequence of dispersion')

    @classmethod
    def get_models(cls):
        return {
            'suggestionlist': DispersionSuggestionlist,
            'scenario': DispersionScenario,
            'treeelement': DispersionTreeElement,
            'event': DispersionEvent,
            'protectionlayer': DispersionProtectionLayer,
            'measure': DispersionMeasure,
            'protectionlayer_measure': DispersionProtectionLayer_Measure,
            'attentionpoint': DispersionAttentionPoint,
            'apsuggestionlist': DispersionAPSuggestionlist,
      }

class DispersionSuggestionlist(base_models.Suggestionlist):
    class Meta:
        permissions = (
            ("display_dispersionsuggestionlists", "May display dispersion suggestion lists"),
            ("edit_dispersionsuggestionlists", "May edit dispersion suggestion lists"),
        )

    appconfig = AppConfig
    displaypermission = appconfig.suggestionlist_displaypermission
    editpermission = appconfig.suggestionlist_editpermission

class DispersionScenario(base_models.Scenario):
    suggestionlist = versioning_models.VersionedForeignKey(DispersionSuggestionlist,
        verbose_name=_('Suggestion list'), blank=True, null=True,
        related_name="scenarios")

    class Meta(base_models.Scenario.Meta):
        ordering = ('type', 'name',)
        permissions = (
            ("display_dispersionscenarios", "May display dispersion scenarios"),
            ("edit_dispersionscenarios", "May edit dispersion scenarios"),
        )

    appconfig = AppConfig
    displaypermission = appconfig.scenario_displaypermission
    editpermission = appconfig.scenario_editpermission

EVENT_ID = base_models.EVENT_ID
PROTECTIONLAYER_ID = base_models.PROTECTIONLAYER_ID

class DispersionTreeElement(base_models.TreeElement):
    scenario = versioning_models.VersionedForeignKey(DispersionScenario, verbose_name = _('Scenario'))
    appconfig = AppConfig


class DispersionEvent(base_models.Event):
    appconfig = AppConfig
    displaypermission = appconfig.scenario_displaypermission
    editpermission = appconfig.scenario_editpermission

class DispersionMeasure(base_models.Measure):
    appconfig = AppConfig
    displaypermission = appconfig.scenario_displaypermission
    editpermission = appconfig.scenario_editpermission

class DispersionProtectionLayer(base_models.ProtectionLayer):
    measures = versioning_models.VersionedManyToManyField(DispersionMeasure,
        verbose_name = _('Measures'),
        through="DispersionProtectionLayer_Measure",
        blank=True, null=True,
        related_name='protectionlayers')

    appconfig = AppConfig
    displaypermission = appconfig.scenario_displaypermission
    editpermission = appconfig.scenario_editpermission

class DispersionProtectionLayer_Measure(base_models.ProtectionLayer_Measure):
    protectionlayer = versioning_models.VersionedForeignKey(DispersionProtectionLayer)
    measure = versioning_models.VersionedForeignKey(DispersionMeasure)

    appconfig = AppConfig
    displaypermission = appconfig.scenario_displaypermission
    editpermission = appconfig.scenario_editpermission

class DispersionAPSuggestionlist(base_models.APSuggestionlist):
    appconfig = AppConfig
    displaypermission = appconfig.suggestionlist_displaypermission
    editpermission = appconfig.suggestionlist_editpermission

class DispersionAttentionPoint(base_models.AttentionPoint):
##    protectionlayer = versioning_models.VersionedForeignKey(DispersionProtectionLayer,
##        verbose_name = _('Protection layer'), blank=True, null=True)
    measure = versioning_models.VersionedForeignKey(DispersionMeasure, verbose_name = _('Measure'),
        blank=True, null=True,
        related_name="attentionpoints")
    suggestionlist = versioning_models.VersionedForeignKey(DispersionAPSuggestionlist,
        verbose_name=_('Suggestion list'), blank=True, null=True,
        related_name="attentionpoints")

    appconfig = AppConfig
    displaypermission = appconfig.scenario_displaypermission
    editpermission = appconfig.scenario_editpermission


make_version = curry(base_models.make_version, appconfig=AppConfig)
versioning_models.post_version.connect(make_version, sender=versioning_models.Version)

analysesmodules.register(AppConfig, 40)

class Dispersions_SuggestionlistSearch(Search):
    permission = 'dispersions.display_dispersionsuggestionlists'
    fields = ['name', 'description']
    model = DispersionSuggestionlist
    name = _('Suggestion list')

searches.register(Dispersions_SuggestionlistSearch)

class Dispersions_ScenarioSearch(Search):
    permission = 'dispersions.display_dispersionscenarios'
    fields = ['name', 'description', 'evaluation']
    model = DispersionScenario
    name = _('Scenario')

searches.register(Dispersions_ScenarioSearch)

class Dispersions_EventSearch(Search):
    permission = 'dispersions.display_dispersionscenarios'
    fields = ['name', 'description']
    model = DispersionEvent
    name = _('Cause')

searches.register(Dispersions_EventSearch)

class Dispersions_ProtectionLayerSearch(Search):
    permission = 'dispersions.display_dispersionscenarios'
    fields = ['name', 'description']
    model = DispersionProtectionLayer
    name = _('Protection layer')

searches.register(Dispersions_ProtectionLayerSearch)

class Dispersions_MeasureSearch(Search):
    permission = 'dispersions.display_dispersionscenarios'
    fields = ['name', 'tag', 'description']
    model = DispersionMeasure
    name = _('Measure')

searches.register(Dispersions_MeasureSearch)

class Dispersions_APSuggestionlistSearch(Search):
    permission = 'dispersions.display_dispersionsuggestionlists'
    fields = ['name', 'description']
    model = DispersionAPSuggestionlist
    name = _('Attention point suggestion list')

searches.register(Dispersions_APSuggestionlistSearch)

class Dispersions_AttentionPointSearch(Search):
    permission = 'dispersions.display_dispersionscenarios'
    fields = ['name', 'description']
    model = DispersionAttentionPoint
    name = _('Attention point')

searches.register(Dispersions_AttentionPointSearch)
