"""
forms for installations app
"""

from analyses.main import analysesmodules

from django import forms
from django.contrib.auth.models import User, Permission
import django.contrib.admin.widgets as adminwidgets
from django.utils.translation import ugettext_lazy as _

from hazards import models as hazards_models

from installations import models as app_models

from utils import forms as utils_forms
from utils import fields as utils_fields
from utils import lists as utils_lists
from utils import widgets as utils_widgets
from utils import sheets
from utils.versioning import versionedforms
from utils.versioning.forms import PublishVersionForm

class SectionForm(versionedforms.VersionedForm):
    class Meta:
        model = app_models.Section
        attrs = {
            'name': {'size':75},
            'description': {'rows':15,'cols':75},
        }
        widgets = {
            'description': utils_widgets.MarkItUpWidget()
        }
        fields = ('name', 'description',)

class SectionSheet(versionedforms.VersionedSheet):
    description = utils_fields.MarkdownField(data = 'description', name='description', label=_('Description'))
    class Meta:
        model = app_models.Section
        fields = ('name', 'description',)

class Section_SubstanceForm(versionedforms.VersionedForm):
    class Meta:
        model = app_models.Section_Substance
        attrs = {
        }
        fields = ('substance', 'aggregation', 'quantity_desired', 'quantity_undesired', 'description',)

    def __init__(self, version=None, *args, **kwargs):
        super(Section_SubstanceForm, self).__init__(*args, **kwargs)
        self.fields['substance'].queryset = hazards_models.Substance.objects.get_versionedobjects(version=version)

class Section_SubstanceSheet(versionedforms.VersionedSheet):
    class Meta:
        model = app_models.Section_Substance
        fields = ('substance', 'aggregation', 'quantity_desired', 'quantity_undesired', 'description',)
        links = ('substance',)

class Section_ReactionForm(versionedforms.VersionedForm):
    class Meta:
        model = app_models.Section_Reaction
        attrs = {
        }
        fields = ('reaction', 'desired', 'description',)

    def __init__(self, version=None, *args, **kwargs):
        super(Section_ReactionForm, self).__init__(*args, **kwargs)
        self.fields['reaction'].queryset = hazards_models.Reaction.objects.get_versionedobjects(version=version)

class Section_ReactionSheet(versionedforms.VersionedSheet):
    class Meta:
        model = app_models.Section_Reaction
        fields = ('reaction', 'desired', 'description',)
        links = ('reaction',)


class InlineWeakpointForm(versionedforms.VersionedForm):
    class Meta:
        model = app_models.Weakpoint
        exclude = ('section',)
    def __init__(self, *args, **kwargs):
        super(InlineWeakpointForm, self).__init__(*args, **kwargs)
        self.fields['name'].widget.attrs = {'size':70}

class InlineWeakpointsFormSet(versionedforms.VersionedInlineFormSet):
    form = InlineWeakpointForm
    model = app_models.Weakpoint
    parent = "section"
    can_delete = True
    extra = 3

class SectionPermissionsForm(versionedforms.VersionedForm):
    authorized_users = forms.ModelMultipleChoiceField(queryset = User.objects.none(),
        required = False,
        help_text = None,
        label = _('Authorized users'),
        widget = adminwidgets.FilteredSelectMultiple(verbose_name=_('Authorized users'), is_stacked=False),
        )
    class Meta:
        model = app_models.Section
        fields = ('authorized_users',)
    def __init__(self, *args, **kwargs):
        super(SectionPermissionsForm, self).__init__(*args, **kwargs)
        self.fields['authorized_users'].queryset = User.objects.exclude(is_superuser=True
            ).exclude(user_permissions=Permission.objects.get_by_natural_key('edit_installations','installations','section')
            ).exclude(groups__permissions=Permission.objects.get_by_natural_key('edit_installations','installations','section'))

class SectionPermissionsSheet(versionedforms.VersionedSheet):
    class Meta:
        model = app_models.Section
        fields = ('authorized_users',)

class ScopeFilterForm(utils_lists.FilterForm):
    """Custom filterform for the scopes list"""
    name = forms.CharField(max_length=100, required=False, label=_('Scope'))

class ScopeForm(versionedforms.VersionedForm):
    class Meta:
        model = app_models.Scope
        attrs = {
            'name': {'size':75},
            'description': {'rows':5,'cols':75},
        }
        fields = ('name', 'description',)

class ScopeSheet(versionedforms.VersionedSheet):
    class Meta:
        model = app_models.Scope
        fields = ('name', 'description',)

def get_analysis_type_choices():
    choices = [('', '---------')]
    choices.extend(analysesmodules.get_analysis_choices())
    return choices

class OverviewFilterForm(utils_lists.FilterForm):
    """Custom filterform for the overview"""
    name = forms.CharField(max_length=100, required=False, label=_('Name'))
    analysistype = forms.ChoiceField(choices=get_analysis_type_choices(), required=False, label=_('Scenario type'))

class AddScenarioForm(forms.Form):
    scope = forms.ChoiceField(choices=(), required=True, label=_('Scope'))
    analysistype = forms.ChoiceField(choices=(), required=True, label=_('Analysis type'))
    use_suggestionlist = forms.BooleanField(required=False, label=_('Use suggestionlist'), widget=forms.CheckboxInput)
    def __init__(self, section=None, scope=None, version=None, user=None, *args, **kwargs):
        self.version = version
        super(AddScenarioForm, self).__init__(*args, **kwargs)
        if section is None:
            del self.fields['scope']
            self.fields['analysistype'].choices = analysesmodules.get_analysis_add_choices(user)
        else:
            self.fields['scope'].choices = [(item.v_id, item.name) for item in section.get_scopes()]
            self.fields['analysistype'].choices = analysesmodules.get_analysis_add_choices(user)

    def clean_scope(self):
        data = self.cleaned_data['scope']
        try:
            scope = app_models.Scope.objects.get_versionedobject(v_id=data, version=self.version)
        except app_models.Scope.DoesNotExist:
            raise forms.ValidationError("Invalid scope!")
        return scope

class OverviewReportForm(forms.Form):
    def __init__(self, user, *args, **kwargs):
        from installations.views.main import get_overview_report_formfields
        super(OverviewReportForm, self).__init__(*args, **kwargs)
        fields = utils_forms.treeify(get_overview_report_formfields(user))
        for k in fields:
            self.fields[k['name']] = k['field']

