"""
Installations app

App to describe the installation layout

"""
from django.contrib.auth.models import User
from django.db import models
from django.db.models import signals as django_signals
from django.utils.translation import ugettext_lazy as _

from hazards import models as hazards_models

from utils.exchange import exchange
from utils import reports as utils_reports
from utils.strings import ugettext_tolanguage as _l
from utils.trees import models as trees_models
from utils.versioning import models as versioning_models
from utils.versioning import versionedmptt

# DENORMALISATION IDEA: store tree_id in section object --> easy to retrieve corresponding tree elements for section


class SectionTreeElement(versionedmptt.VersionedMpttModel):
    section = versioning_models.VersionedForeignKey("Section", verbose_name = _('Section'))

    def __unicode__(self):
        if self.section:
            return self.section.__unicode__()
        return u'Unbound SectionTreeElement'


    # method overrides

    def delete(self, user, *args, **kwargs):
        "remove leaf when deleting the treeelement"
        nocascade = kwargs.pop('nocascade', False)
        super(SectionTreeElement, self).delete(user, *args, **kwargs)
        if not nocascade:
            self.section.delete(user, *args, **kwargs)
    delete.alters_data = True


    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def may_change(cls, user):
        if user.has_perm('installations.edit_installations'):
            return True
        return cls.get_authorized_tree(user).count() > 0

    @classmethod
    def get_authorized_tree(cls, user):
        if user.has_perm('installations.edit_installations'):
            return SectionTreeElement.tree.all()
        else:
            s = Section.objects.filter(authorized_users=user)
            if s:
                tree_ids = SectionTreeElement.objects.filter(parent__isnull=True, section__in=s).values_list('tree_id', flat=True)
                if tree_ids:
                    return SectionTreeElement.tree.filter(tree_id__in=tree_ids)
        return SectionTreeElement.objects.none()

    # urls

    def get_absolute_url(self):
        return self.section.get_absolute_url()

    # custom methods

    @classmethod
    def collect_items(cls, version):
        return list((_('Section tree'), item, None) for item in cls.objects.filter(version=version, parent__isnull=True, deleted=False))

    def collect_objects(self, filterkwarg={}):
        """
        Collect the child objects that satisfy the filterkwarg
        Returns a list of lists
        """
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion

        objects = []

        # section
        objects.extend(self.section.collect_objects(filterkwarg))

        # only add self if filterkwarg satisfied:
        if filterkwarg:
            selflist = list(self.__class__.objects.filter(pk=self.pk).filter(**filterkwarg).get_versionedobjects(version=version))
            objects.append(selflist)
        else:
            objects.append([self])

        return objects

    def publish(self, user, instance, include_descendants=False):
        """
        Create a published version of a section
        """
        objects = []
        if include_descendants:
            sectiontreeelements = self.get_descendants(include_self=True)
            for sectiontreeelement in sectiontreeelements:
                objects.extend(sectiontreeelement.collect_objects(filterkwarg={'published_dirty': True}))
        else:
            objects = self.collect_objects(filterkwarg={'published_dirty': True})
        instance.make_version_for_objects(objects)

class Section (versioning_models.VersionedModel):
    name = models.CharField(verbose_name=_('Name'), max_length=100)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    substances = versioning_models.VersionedManyToManyField(hazards_models.Substance, through='Section_Substance')
    reactions = versioning_models.VersionedManyToManyField(hazards_models.Reaction, through='Section_Reaction')
    authorized_users = models.ManyToManyField(User, verbose_name=_('Authorized users'))

    class Meta:
        verbose_name = _('Section')
        verbose_name_plural = _('Sections')
        permissions = (
            ("display_installations", "May display installation information"),
            ("edit_installations", "May edit installation information"),
        )

    def __unicode__(self):
        return self.name


    # method overrides

    save_accepts_raw=True
    def save(self, user, raw=False, *args, **kwargs):
        """
        custom save that adds an autoscope and modifies the autoscope name to match the section's name
        When adding a new object, also create a tree element.
        Optionaly pass a parent kwarg to determine the parent element in the tree
        Returns the tree element if created
        """
        treeparent = kwargs.pop('parent', None)
        new = (not self.pk)
        super(Section, self).save(user, *args, **kwargs)
        if new and not raw:
            scope = Scope(
                name = self.name,
                auto=True
            )
            scope.save(user)
            inst = Scope_Section(
                scope = scope,
                section = self
                )
            inst.save(user)
            treeelement = SectionTreeElement()
            treeelement.section = self
            treeelement.parent = treeparent
            treeelement.save(user)
            return treeelement
        else:
            scopes = [item.scope for item in self.scope_section_set.select_related_versioned('scope')]
            for scope in scopes:
                if scope.auto:
                    if scope.name != self.name:
                        scope.name = self.name
                        scope.save(user)
    save.alters_data = True

    def delete(self, user, *args, **kwargs):
        """
        remove section from scopes, clean up autoscope
        also set item in tree deleted
        """
        if self.version:
            raise Exception('Forbidden to delete versioned element')
        else:
            # remove weakpoints first
            for item in self.weakpoint_set.all():
                item.delete(user)

            # remove membership of scopes and delete autoscope
            sss = self.scope_section_set.select_related_versioned('scope')
            for ss in sss:
                if ss.scope.auto:
                    ss.scope.delete(user,*args, **kwargs)
                ss.delete(user, *args, **kwargs)

            # base delete
            super(Section, self).delete(user, *args, **kwargs)

            # remove treeelement
            sss = self.sectiontreeelement_set.all()
            for ss in sss:
                ss.delete(user, nocascade=True, *args, **kwargs)

    delete.alters_data = True


    # urls

    @models.permalink
    def get_absolute_url(self, tab='scenarios'):
        return ('installations_section', (self.v_id, tab))

    @models.permalink
    def get_edit_url(self, tab='section'):
        return ('installations_section_edit', (self.v_id, tab))

    @models.permalink
    def get_delete_url(self):
        return ('installations_section_delete', (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('installations_section_report', (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('installations_section_compare', (self.v_id, version.id))

    @models.permalink
    def get_publish_url(self):
        return ('installations_section_publish', (self.v_id,))

    @models.permalink
    def get_add_section_url(self):
        return ('installations_section_add_section', (self.v_id,))

    @models.permalink
    def get_add_substance_url(self):
        return ('installations_section_add_substance', (self.v_id,))

    @models.permalink
    def get_add_reaction_url(self):
        return ('installations_section_add_reaction', (self.v_id,))

    @models.permalink
    def get_add_scenario_url(self):
        return ('installations_section_add_scenario', (self.v_id,))


    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def may_add(cls, user, parent=None):
        if parent:
            return parent.may_change(user)
        return user.has_perm('installations.edit_installations')

    def may_change(self, user):
        if user.has_perm('installations.edit_installations'):
            return not self.version
        if not self.is_root_node():
            return self.get_root().may_change(user)
        else:
            return user in self.authorized_users.all()
        return False

    def may_delete(self, user):
        """
        Only allow delete if section is a leaf and has no weakpoints
        no scenarios in autoscope
        no section_substances, no section_reactions
        """
# TODO: this breaches the module logic!
        from analyses.main import count_scenarios_for_scope
        if self.may_change(user):
            if self.is_leaf_node():
                if (self.weakpoint_set.count() == 0 and
                        self.section_substance_set.count() == 0 and
                        self.section_reaction_set.count() == 0):
                    version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
                    return count_scenarios_for_scope(self.get_auto_scope(), version) == 0
        return False

    def may_publish(self, user):
        """
        Only allow publishing when section published in tree
        """
        if self.version:
            return False
        if versioning_models.Version.may_add(user):
            # check tree published
            el = self.get_treeelement()
            if el.__class__.objects.filter(v_id=el.v_id, published=True).count() == 0:
                return False
            # check substances and reactions published
            if Section_Substance.objects.filter(section=self, published=True).count() == 0:
                return False
            if Section_Reaction.objects.filter(section=self, published=True).count() == 0:
                return False
            return True
        return False

    # custom methods

    def get_auto_scope(self):
        sss = self.scope_section_set.select_related_versioned('scope')
        for ss in sss:
            if ss.scope.auto:
                return ss.scope

    def get_treeelement(self):
        if not hasattr(self, '_treeelement'):
            self._treeelement = self.sectiontreeelement_set.all()[0]
        return self._treeelement

    def get_scopes(self):
        "aggregate all scopes from higher sections"
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        sections = self.get_ancestors(include_self=True)
        if sections:
            ss = list(Scope_Section.objects.filter(section__in=sections
                ).get_versionedobjects(version).values_list('scope_id', flat=True))
            if ss:
                return Scope.objects.filter(v_id__in=ss).get_versionedobjects(version)
        return Scope.objects.none()

    def get_descendants(self, include_self=False):
        cachename = include_self and '_descendants_is' or '_descendants'
        if not hasattr(self, cachename):
            el = self.get_treeelement()
            if el.is_leaf_node():
                if include_self:
                    return [self]
                return []
            treeelements = el.get_descendants(include_self=include_self).get_versionedobjects(version=el._version).select_related_versioned()
            res = [element.section for element in treeelements]
            setattr(self, cachename, res)
        return getattr(self, cachename)

    def get_ancestors(self, include_self=False):
        cachename = '_ancestors'
        if not hasattr(self, cachename):
            el = self.get_treeelement()
            if el.is_root_node():
                if include_self:
                    return [self]
                return []
            treeelements = el.get_ancestors().get_versionedobjects(version=el._version).select_related_versioned()
            res = [element.section for element in treeelements]
            setattr(self, cachename, res)
        res = getattr(self, cachename)
        if include_self:
            res.append(self)
        return res

    def get_root(self):
        el = self.get_treeelement()
        return el.get_root().section

    def is_leaf_node(self):
        try:
            return self.get_treeelement().is_leaf_node()
        except:
            return True

    def is_root_node(self):
        try:
            return self.get_treeelement().is_root_node()
        except:
            return False

    def get_section_substances(self):
        """
        aggregate all substances from lower sections
        returns the section_substance objects
        """
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        sections = self.get_descendants(include_self=True)
        if sections:
            return Section_Substance.objects.filter(section__in=sections
                ).get_versionedobjects(version
                ).order_by('substance', 'section').select_related_versioned()
        return Section_Substance.objects.none()

    def get_section_reactions(self):
        """
        aggregate all reactions from lower sections
        returns the section_reaction objects
        """
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        sections = self.get_descendants(include_self=True)
        if sections:
            return Section_Reaction.objects.filter(section__in=sections
                ).get_versionedobjects(version
                ).order_by('reaction', 'section').select_related_versioned()
        return Section_Reaction.objects.none()


    def get_substances_and_reactions(self):
        """
        aggregate all possible reactions between the substances from this and
        lower sections
        returns two querysets: substances and reactions
        """
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        sections = self.get_descendants(include_self=True)
        substances = hazards_models.Substance.objects.none()
        reactions = hazards_models.Reaction.objects.none()
        if sections:
            ss = list(Section_Substance.objects.filter(section__in=sections
                ).get_versionedobjects(version).values_list('substance_id',flat=True))
            if ss:
                substances = hazards_models.Substance.objects.filter(models.Q(v_id__in=ss)|models.Q(universal=True)
                    ).get_versionedobjects(version)
                if substances:
                    # get reactions for these substances
                    reactions = hazards_models.Reaction.objects.filter(models.Q(substance1__in=substances)|models.Q(substance2__in=substances)
                        ).get_versionedobjects(version
                        ).select_related_versioned()
        return substances, reactions

    def get_possible_reactions(self):
        """
        aggregate all possible reactions between the substances from this and
        lower sections
        returns reaction objects
        """
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        sections = self.get_descendants(include_self=True)
        substances = hazards_models.Substance.objects.none()
        reactions = hazards_models.Reaction.objects.none()
        if sections:
            ss = list(Section_Substance.objects.filter(section__in=sections
                ).get_versionedobjects(version).values_list('substance_id',flat=True))
            if ss:
                substances = list(hazards_models.Substance.objects.filter(models.Q(v_id__in=ss)|models.Q(universal=True)
                    ).get_versionedobjects(version).values_list('id', flat=True))
                if substances:
                    # get reactions for these substances
                    reactions = hazards_models.Reaction.objects.filter(models.Q(substance1__in=substances)|models.Q(substance2__in=substances)
                        ).get_versionedobjects(version
                        ).select_related_versioned()
        return reactions

    def collect_objects(self, filterkwarg={}):
        """
        Collect the child objects that satisfy the filterkwarg
        Returns a list of lists
        """
        from analyses.main import collect_objects_for_scope
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion

        objects = []

        # only add self if filterkwarg satisfied:
        if filterkwarg:
            selflist = list(self.__class__.objects.filter(pk=self.pk).filter(**filterkwarg).get_versionedobjects(version=version))
            objects.append(selflist)
        else:
            objects.append([self])

        # substance and reaction links
        objects.append(list(Section_Substance.objects.filter(section=self).filter(**filterkwarg).get_versionedobjects(version=version)))
        objects.append(list(Section_Reaction.objects.filter(section=self).filter(**filterkwarg).get_versionedobjects(version=version)))

        # add autoscope and its scenarios
        for ss in self.scope_section_set.select_related_versioned('scope'):
            if ss.scope.auto:
                objects.extend(ss.scope.collect_objects(filterkwarg))
                if filterkwarg:
                    objects.append(list(Scope_Section.objects.filter(v_id=ss.v_id).filter(**filterkwarg).get_versionedobjects(version=version)))
                else:
                    objects.append([ss])

        return objects

    def publish(self, user, instance, include_descendants=False):
        """
        Create a published version of a section
        """
        objects = []
        if include_descendants:
            el = self.get_treeelement()
            el.publish(user, instance, include_descendants=True)
        else:
            objects = self.collect_objects(filterkwarg={'published_dirty': True})
        instance.make_version_for_objects(objects)

    def dump(self, include_descendants=False):
        """
        Create a json dump of this section
        """
        objects = []
        if include_descendants:
            # include tree elements too
            el = self.get_treeelement()
            sectiontreeelements = el.get_descendants(include_self=True)
            for sectiontreeelement in sectiontreeelements:
                for l in sectiontreeelement.collect_objects(filterkwarg={}):
                    objects.extend(l)
        else:
            for l in self.collect_objects(filterkwarg={}):
                objects.extend(l)
        return exchange.dump(objects)

    @classmethod
    def get_report_formfields(cls, user):
        from django import forms
        from analyses.main import analysesmodules
        res = [{'name': 'subsections',
            'field': forms.BooleanField(required=False, label=_('Subsections'),),
            'children': []}]
        res.extend(analysesmodules.get_report_formfields(user))
        return res

    def get_report_body(self, user, language=u'en', choices={}, subsection=False):
        from installations import forms as app_forms
        from installations import tables as app_tables
        from analyses.main import get_scenarios_for_scopes, get_scenarios_for_scope, build_scenariotree
        form = app_forms.SectionSheet(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form, 'title': u'%s: %s' % (_l('Section', language), self)}
        el = self.sectiontreeelement_set.all()[0]
        subsections = el.get_descendants(include_self=True).get_versionedobjects(version=el._version).select_related_versioned()
        if len(subsections) > 1:
            context['subsections'] = subsections

        tables = []
        queryset = self.weakpoint_set.all()
        if queryset:
            table = app_tables.WeakpointsTable(queryset)
            table.title = _l('Weak points', language)
            tables.append(table)
        # aggregate all substances from lower sections
        queryset = self.get_section_substances()
        if queryset:
            table = app_tables.Section_SubstancesTable(queryset)
            table.title = _l('Substances', language)
            tables.append(table)
        queryset = self.get_section_reactions()
        if queryset:
            table = app_tables.Section_ReactionsTable(queryset)
            table.title = _l('Reactions', language)
            tables.append(table)
        context['tables'] = tables

        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        scenarios = get_scenarios_for_scopes(self.get_scopes(), user, version)
        context['scenarios'] = build_scenariotree(scenarios, scope=self.get_auto_scope())
        html = [utils_reports.make_pdf_body(context, language=language, template='installations/section_report.html')]
        if choices.get('scenarios'):
            if subsection:
                scenarios = get_scenarios_for_scope(self.get_auto_scope(), user, version)
            for mod in scenarios:
                for scenario in mod['items']:
                    html.append('<pdf:nextpage>')
                    html.extend(scenario.get_report_body(user, language, choices))
        if choices.get('subsections'):
            for s in subsections:
                if s.section != self:
                    html.append('<pdf:nextpage>')
                    html.extend(s.section.get_report_body(user, language, choices, subsection=True))
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}):
        from analyses.main import analysesmodules
        css_files = []
        for mod in analysesmodules.get_modules():
            css_files.extend(mod.get_model('scenario').get_report_css(language, choices))
        return css_files

    def get_report(self, user, language=u'en', choices={}):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

# Explicit m2m model
class Section_Substance(versioning_models.VersionedModel):
    section = versioning_models.VersionedForeignKey(Section, verbose_name = _('Section'))
    substance = versioning_models.VersionedForeignKey(hazards_models.Substance, verbose_name = _('Substance'))
    aggregation = models.CharField(verbose_name = _('Aggregation'), max_length=30, blank=True, null=True)
    quantity_desired = models.CharField(verbose_name = _('Quantity - desired'), max_length=50, blank=True, null=True)
    quantity_undesired = models.CharField(verbose_name = _('Quantity - undesired'), max_length=50, blank=True, null=True)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)

    class Meta:
        verbose_name = _('Substance for section')
        verbose_name_plural = _('Substances for section')

    def __unicode__(self):
        return '%s in %s' % (self.substance, self.section)


    # urls

    @models.permalink
    def get_absolute_url(self):
        return ('installations_section_substance', (self.v_id,))

    @models.permalink
    def get_edit_url(self):
        return ('installations_section_substance_edit', (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('installations_section_substance_delete', (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('installations_section_substance_compare', (self.v_id, version.id))


    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def may_add(cls, user, parent):
        return parent.may_change(user)

    def may_change(self, user):
        return self.section.may_change(user)

    def may_delete(self, user):
        return self.section.may_change(user)


    # custom methods

    @classmethod
    def collect_items(cls, version):
        modified = list(cls.objects.filter(version=version).values_list('section_id', flat=True).distinct())
        if modified:
            items = Section.objects.filter(v_id__in=modified).get_versionedobjects(version=version)
            return list(('Substances for section', item, None) for item in items)
        return []

# Explicit m2m model
class Section_Reaction(versioning_models.VersionedModel):
    section = versioning_models.VersionedForeignKey(Section, verbose_name = _('Section'))
    reaction = versioning_models.VersionedForeignKey(hazards_models.Reaction, verbose_name = _('Reaction'))
    desired = models.BooleanField(verbose_name = _('Desired'))
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)

    class Meta:
        verbose_name = _('Reaction for section')
        verbose_name_plural = _('Reactions for section')

    def __unicode__(self):
        return '%s in %s' % (self.reaction, self.section)


    # urls

    @models.permalink
    def get_absolute_url(self):
        return ('installations_section_reaction', (self.v_id,))

    @models.permalink
    def get_edit_url(self):
        return ('installations_section_reaction_edit', (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('installations_section_reaction_delete', (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('installations_section_reaction_compare', (self.v_id, version.id))


    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def may_add(cls, user, parent):
        return parent.may_change(user)

    def may_change(self, user):
        return self.section.may_change(user)

    def may_delete(self, user):
        return self.section.may_change(user)


    # custom methods

    @classmethod
    def collect_items(cls, version):
        modified = list(cls.objects.filter(version=version).values_list('section_id', flat=True).distinct())
        if modified:
            items = Section.objects.filter(v_id__in=modified).get_versionedobjects(version=version)
            return list(('Reactions for section', item, None) for item in items)
        return []


class Weakpoint(versioning_models.VersionedModel):
    section = versioning_models.VersionedForeignKey(Section, verbose_name = _('Section'))
    name = models.CharField(verbose_name=_('Name'), max_length=100)

    class Meta:
        verbose_name = _('Weak point')
        verbose_name_plural = _('Weak points')
        ordering = ('name',)

    def __unicode__(self):
        return self.name


    # permissions

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def may_add(cls, user, parent):
        return parent.may_change(user)

    def may_change(self, user):
        return self.section.may_change(user)

    def may_delete(self, user):
        return self.section.may_change(user)


    # custom methods

    @classmethod
    def collect_items(cls, version):
        """
        Aggregating modifications to weakpoints per section
        """
        modified = list(cls.objects.filter(version=version).values_list('section_id', flat=True).distinct())
        if modified:
            items = Section.objects.filter(v_id__in=modified).get_versionedobjects(version=version)
            return list(('Weakpoints for section', item, None) for item in items)
        return []

class Scope(versioning_models.VersionedModel):
    name = models.CharField(verbose_name=_('Name'), max_length=100)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    sections = versioning_models.VersionedManyToManyField(Section, through = "Scope_Section", related_name="scopes")
    auto = models.BooleanField(verbose_name = _('Automatic'), default=False)

    class Meta:
        verbose_name = _('Scope')
        verbose_name_plural = _('Scopes')
        ordering = ('name',)

    def __unicode__(self):
        return self.name


    # method overrides

    def delete(self, user, *args, **kwargs):
        "set m2m's deleted too"
        if not self.version:
            # remove memberships
            for ss in self.scope_section_set.all():
                ss.delete(user, *args, **kwargs)
            super(Scope, self).delete(user, *args, **kwargs)
    delete.alters_data = True


    # urls

    @models.permalink
    def get_absolute_url(self, tab='scenarios'):
        return ('installations_scope', (self.v_id, tab))

    @models.permalink
    def get_edit_url(self, tab='scope'):
        return ('installations_scope_edit', (self.v_id, tab))

    @models.permalink
    def get_delete_url(self):
        return ('installations_scope_delete', (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('installations_scope_report', (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('installations_scope_compare', (self.v_id, version.id))

    @models.permalink
    def get_publish_url(self):
        return ('installations_scope_publish', (self.v_id,))

    @models.permalink
    def get_add_scenario_url(self):
        return ('installations_scope_add_scenario', (self.v_id,))

    def get_auto_section(self):
        if self.auto:
            if not hasattr(self, '_auto_section'):
                self._auto_section = self.scope_section_set.all()[0].section
            return self._auto_section
        return None


    def get_redirect_url(self):
        if self.auto:
            return self.get_auto_section().get_absolute_url()
        return self.get_absolute_url()


    # permissions

# TODO ? do these need to be linked to section permissions ?
# or only allow users that have permissions on the entire installation to define custom scopes

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def may_add(cls, user):
        return user.has_perm('installations.edit_installations')

    def may_change(self, user):
        if self.auto:
            return False
        if user.has_perm('installations.edit_installations'):
            return not self.version
        return False

    def may_delete(self, user):
        """
        Only allow delete if section is a leaf and has no weakpoints
        no scenarios in autoscope
        no section_substances, no section_reactions
        """
# TODO: this breaches the module logic!
        from analyses.main import count_scenarios_for_scope
        if self.may_change(user):
            version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
            return count_scenarios_for_scope(self, version) == 0
        return False

    def may_publish(self, user):
        """
        Only allow publishing when all sections published
        """
        if self.version:
            return False
        if self.auto:
            return False
        if versioning_models.Version.may_add(user):
            for item in self.scope_section_set.all():
                if Section.objects.filter(v_id=item.section.v_id, published=True).count() == 0:
                    return False
            return True
        return False
    # custom methods

    def get_breadcrumbs(self):
        if self.auto:
            item = self.get_auto_section()
            return [{'url': item.get_absolute_url(), 'label': u'%s' % item}]
        return [{'url': self.get_absolute_url(), 'label': u'%s' % self}]


    def collect_objects(self, filterkwarg={}):
        """
        Collect the child objects that satisfy the filterkwarg
        Returns a list of lists
        """
        from analyses.main import collect_objects_for_scope
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion

        objects = []

        # only add self if filterkwarg satisfied:
        if filterkwarg:
            selflist = list(self.__class__.objects.filter(pk=self.pk).filter(**filterkwarg).get_versionedobjects(version=version))
            objects.append(selflist)
        else:
            objects.append([self])

        # scenarios
        objects.extend(collect_objects_for_scope(self, version, filterkwarg))

        return objects

    def publish(self, user, instance):
        """
        Create a published version of a scope
        """
        objects = self.collect_objects(filterkwarg={'published_dirty': True})
        instance.make_version_for_objects(objects)

    def dump(self):
        """
        Create a json dump of this scope
        """
        objects = []
        for item in self.collect_objects(filterkwarg={}):
            objects.extend(item)
        return exchange.dump(objects)

    @classmethod
    def collect_items(cls, version):
        """
        Collect all items that were changed in a version.
        Exclude auto scopes.
        """
        objects = cls.objects.filter(version=version, auto=False)
        res = []
        for item in objects:
            change = None
            if item.new:
                change = _('New')
            if item.deleted:
                change = _('Deleted')
            res.append((_('Scope'), item, change))
        return res

    @classmethod
    def get_listreport_formfields(cls, user):
        from django import forms
        from analyses.main import analysesmodules
        res = [{'name': 'scopes',
            'field': forms.BooleanField(required=False, label=_('Scope sheets'),),
            'children': cls.get_report_formfields(user)}]
        return res

    @classmethod
    def get_report_formfields(cls, user):
        from analyses.main import analysesmodules
        return analysesmodules.get_report_formfields(user)

    def get_report_body(self, user, language=u'en', choices={}):
        from installations import forms as app_forms
        from analyses.main import get_scenarios_for_scope, build_scenariotree
        form = app_forms.ScopeSheet(instance=self, omit_empty=True)
        context = {'object': self, 'form': form, 'title': u'%s: %s' % (_l('Scope', language), self)}
        version = getattr(self, '_version', self.version) or versioning_models.WorkVersion
        scenarios = get_scenarios_for_scope(self, user, version)
        context['scenarios'] = build_scenariotree(scenarios, scope=self)
        html = [utils_reports.make_pdf_body(context, language=language, template='installations/section_report.html')]
        if choices.get('scenarios'):
            for mod in scenarios:
                for scenario in mod['items']:
                    html.append('<pdf:nextpage>')
                    html.extend(scenario.get_report_body(user, language, choices))
        return html

    @classmethod
    def get_report_css(cls, language=u'en', choices={}):
        from analyses.main import analysesmodules
        css_files = []
        for mod in analysesmodules.get_modules():
            css_files.extend(mod.get_model('scenario').get_report_css(language, choices))
        return css_files

    def get_report(self, user, language=u'en', choices={}):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

# Explicit m2m model to allow versioning
class Scope_Section(versioning_models.VersionedModel):
    scope = versioning_models.VersionedForeignKey(Scope)
    section = versioning_models.VersionedForeignKey(Section)

    def __unicode__(self):
        return '%s in %s' % (self.section, self.scope)

    @classmethod
    def may_display(cls, user):
        return user.has_perm('installations.display_installations')

    @classmethod
    def collect_items(cls, version):
        modified = list(cls.objects.filter(version=version).values_list('scope_id', flat=True).distinct())
        if modified:
            items = Scope.objects.filter(v_id__in=modified).filter(auto=False).get_versionedobjects(version=version)
            return list(('Sections for scope', item, None) for item in items)
        return []

##def check_foreignkey_published():
##    valid = True
##    errors = []
##    filterkwarg={'published_dirty': True, 'version__isnull': True}
##    for item in Section_Substance.objects.filter(**filterkwarg):
##        v, e = item.check_foreignkey_published(['substance'])
##        if not v:
##            valid = False
##            errors.extend(e)
##    for item in Section_Reaction.objects.filter(**filterkwarg):
##        v, e = item.check_foreignkey_published(['reaction'])
##        if not v:
##            valid = False
##            errors.extend(e)
##    return valid, errors

def may_publish(user):
    if user.has_perm('installations.edit_installations') and user.has_perm('versioning.edit_versions'):
        # make sure each linked substance or reaction has at least one published version
        # using extra to avoid high querycount
        qs = Section_Substance.objects.filter(version__isnull=True).extra(
            where=[
                '''(SELECT COUNT(*) FROM hazards_substance
                    WHERE hazards_substance.v_id = installations_section_substance.substance_id
                    and hazards_substance.published) = 0
                ''',
            ],
        )
        if qs:
            return False
        qs = Section_Reaction.objects.filter(version__isnull=True).extra(
            where=[
                '''(SELECT COUNT(*) FROM hazards_reaction
                    WHERE hazards_reaction.v_id = installations_section_reaction.reaction_id
                    and hazards_reaction.published) = 0
                ''',
            ],
        )
        if qs:
            return False
        return True
    return False

def publish_installations(user, instance):
    """
    Create a published version of the installations app
    """

    objects = []

    filterkwarg={'published_dirty': True, 'version__isnull': True}

    # sections
    objects.append(list(Section.objects.filter(**filterkwarg)))

    # section trees
    objects.append(list(SectionTreeElement.tree.root_nodes().filter(**filterkwarg)))

    # section_substance
    objects.append(list(Section_Substance.objects.filter(**filterkwarg)))

    # section_reaction
    objects.append(list(Section_Reaction.objects.filter(**filterkwarg)))

    # weakpoints
    objects.append(list(Weakpoint.objects.filter(**filterkwarg)))

    # scopes
    objects.append(list(Scope.objects.filter(**filterkwarg)))

    # scope_section
    objects.append(list(Scope_Section.objects.filter(**filterkwarg)))

    instance.make_version_for_objects(objects)


def make_installations_version(sender, instance, objects, **kwargs):
    """
    make version of dirty objects
    """
    if not objects:
        # only version everything when objects are not specifically included
        if instance.published:
            filterkwarg={'published_dirty': True, 'version__isnull': True}
        else:
            filterkwarg={'dirty': True, 'version__isnull': True}
        for item in Section.objects.filter(**filterkwarg):
            item.make_version(instance)
        for item in SectionTreeElement.tree.root_nodes().filter(**filterkwarg):
            item.make_version(instance)
        for item in Section_Substance.objects.filter(**filterkwarg):
            item.make_version(instance)
        for item in Section_Reaction.objects.filter(**filterkwarg):
            item.make_version(instance)
        for item in Weakpoint.objects.filter(**filterkwarg):
            item.make_version(instance)
        for item in Scope.objects.filter(**filterkwarg):
            item.make_version(instance)
        for item in Scope_Section.objects.filter(**filterkwarg):
            item.make_version(instance)

versioning_models.post_version.connect(make_installations_version, sender=versioning_models.Version)


from utils.search.models import searches, Search

class Installations_SectionSearch(Search):
    permission = 'installations.display_installations'
    fields = ['name', 'description']
    model = Section

    def get_url(self, object):
        return object.get_absolute_url('section')

searches.register(Installations_SectionSearch)

class Installations_WeakpointSearch(Search):
    permission = 'installations.display_installations'
    fields = ['name', ]
    model = Weakpoint
    select_related_versioned = True

    def get_url(self, object):
        return object.section.get_absolute_url('weakpoints')

searches.register(Installations_WeakpointSearch)

class Installations_ScopeSearch(Search):
    permission = 'installations.display_installations'
    fields = ['name', 'description']
    model = Section
    select_related_versioned = True

searches.register(Installations_SectionSearch)

