from django.contrib.auth.models import User, AnonymousUser
from django.test import TestCase

from hazards import models as hazards_models

from installations import models as app_models

from mptt.utils import tree_item_iterator

import time

from utils import tests as utils_tests
from utils.trees import models as trees_models
from utils.trees import views as trees_views
from utils.versioning.models import Version, WorkVersion
from utils.versioning.helpers import annotate_versioned


class InstallationsTestCase(TestCase):
    fixtures = ['dummyusers', 'versioning', 'hazards', 'installations']

    def setUp(self):
        from django.contrib.auth.models import User, Permission
        from django.contrib.contenttypes.models import ContentType
        # give the users permissions

        # authorisations for users:
        # 1 = authorised
        # 2 = unauthorised to edit
        # 3 = unauthorised to even display

        u = User.objects.get(pk=1)

        permissions = [
            ('display_versions','versioning','version'),
            ('edit_versions','versioning','version'),
            ('display_installations','installations','section'),
            ('edit_installations','installations','section'),
            ('display_deviationscenarios','deviations','deviationscenario'),
            ('edit_deviationscenarios','deviations','deviationscenario'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

        u = User.objects.get(pk=2)
        permissions = [
            ('display_versions','versioning','version'),
            ('display_installations','installations','section'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

class InstallationsSectionTestCase(InstallationsTestCase):
    def test_section_views(self):

        base_url = '/installations/sections/'
        overview_url = '/installations/overview/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3)

        class installations(utils_tests.TestDefinition):
            description = 'main installations view'
            url = '/installations/'
            users = [1,2,]

        class overview(utils_tests.TestDefinition):
            description = 'installations overview'
            url = overview_url
            users = [1,2,]
            contains = [{'url': overview_url + 'edit/', 'users': [1,]},
            # add no longer on overview but in edit
                {'url': base_url + 'add/', 'users': []},
            # no publish because the hazards info is not published
                {'url': '/installations/publish/', 'users': []}]

        class overview_edit(utils_tests.TestDefinition):
            # this is only the get test
            description = 'installations overview edit'
            url = overview_url + 'edit/'
            users = [1,]

        class overview_report(utils_tests.edit_TestDefinition):
            description = 'installations overview report'
            url = overview_url + 'pdf/'
            users = [1,2,]
            data = {
                "sections": 1,
                "submit_pdf": "Generate pdf",
            }

        class overview_compare(utils_tests.TestDefinition):
            description = 'installations overview compare'
            url = overview_url + 'compare/0/'
            users = [1,2,]

        class overview_history(utils_tests.TestDefinition):
            description = 'installations history'
            url = overview_url + 'history/'
            users = [1,2,]

        class overview_tree_json(utils_tests.TestDefinition):
            description = 'get json tree part'
            url = '/installations/tree/json/'
            data = {'id': 1}
            users = [1,2,]
            contains = [{'url': base_url + '2/', 'users': [1,2]}]

        class section1_section(utils_tests.TestDefinition):
            description = 'display section 1 tab section'
            # section 1 has subsections: no delete
            url = base_url + '1/section/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/section/edit/', 'users': [1,]},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class section2_section(utils_tests.TestDefinition):
            description = 'display section 2 tab section'
            # section 2 has no subsections: delete allowed
            url = base_url + '2/section/'
            users = [1,2,]
            contains = [
                {'url': base_url + '2/section/edit/', 'users': [1,]},
                {'url': base_url + '2/delete/', 'users': [1,]},
            ]

        class section_weakpoints(utils_tests.TestDefinition):
            description = 'display section tab weakpoints'
            url = base_url + '1/weakpoints/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/weakpoints/edit/', 'users': [1]},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class section_subsections(utils_tests.TestDefinition):
            description = 'display section tab subsections'
            url = base_url + '1/subsections/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/subsections/edit/', 'users': [1,]},
            ]

        class section_substances(utils_tests.TestDefinition):
            description = 'display section tab substances'
            url = base_url + '1/substances/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/substances/edit/', 'users': []},
                {'url': base_url + '1/delete/', 'users': []},
                {'url': base_url + '1/substance/add/', 'users': [1,]},
            ]

        class section_reactions(utils_tests.TestDefinition):
            description = 'display section tab reactions'
            url = base_url + '1/reactions/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/reactions/edit/', 'users': []},
                {'url': base_url + '1/delete/', 'users': []},
                {'url': base_url + '1/reaction/add/', 'users': [1,]},
            ]

        class section_scenarios(utils_tests.TestDefinition):
            description = 'display section tab scenarios'
            url = base_url + '1/scenarios/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/scenarios/edit/', 'users': []},
                {'url': base_url + '1/delete/', 'users': []},
                {'url': base_url + '1/scenario/add/', 'users': [1,]},
            ]

        class section1_permissions(utils_tests.TestDefinition):
            description = 'display section 1 tab permissions'
            url = base_url + '1/permissions/'
            users = [1,]
            contains = [
                {'url': base_url + '1/permissions/edit/', 'users': [1,]},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class section2_permissions(utils_tests.TestDefinition):
            description = 'display section 2 tab permissions'
            url = base_url + '2/permissions/'
            # This should not be available to anyone
            users = []
            contains = [
                {'url': base_url + '2/permissions/edit/', 'users': []},
                {'url': base_url + '2/delete/', 'users': []},
            ]

        class section_report(utils_tests.edit_TestDefinition):
            description = 'pdf report for section'
            url = base_url + '1/pdf/'
            users = [1,2,]
            data = {
                "submit_pdf": "Generate pdf",
            }

        class section_compare(utils_tests.TestDefinition):
            description = 'compare a section'
            url = base_url + '1/compare/0/'
            users = [1,2,]

        class section_add(utils_tests.add_TestDefinition):
            description = 'add a section'
            url = base_url + 'add/'
            users = [1,]
            data = {
                "name": "bogus",
                "description": "",
                "management-referer": overview_url
            }
            redirect_url = overview_url
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section.objects.order_by('-id').get_versionedobjects(version=WorkVersion)[0].delete(testcase.user, override=True)

        class subsection_add(utils_tests.add_TestDefinition):
            description = 'add a subsection to a section'
            url = base_url + '1/section/add/'
            users = [1,]
            data = {
                "name": "bogus",
                "description": "",
                "management-referer": base_url + '1/section/'
            }
            redirect_url = base_url + '1/section/'
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section.objects.order_by('-id')[0].delete(testcase.user, override=True)

        class section_edit(utils_tests.edit_TestDefinition):
            description = 'edit a section'
            url = base_url + '1/section/edit/'
            users = [1,]
            data = {
                "name": "A1",
                "description": "",
                "management-referer": base_url + '1/section/'
            }
            redirect_url = base_url + '1/section/'

        class section_weakpoints_edit(utils_tests.edit_TestDefinition):
            description = 'edit the weakpoints for a section, adding one'
            url = base_url + '1/weakpoints/edit/'
            users = [1,]
            data = {
                "weakpoint_set-TOTAL_FORMS": "5",
                "weakpoint_set-INITIAL_FORMS": "2",
                "weakpoint_set-MAX_NUM_FORMS": "0",
                "weakpoint_set-0-id": 1,
                "weakpoint_set-0-name": "Drainpunt",
                "weakpoint_set-1-id": 2,
                "weakpoint_set-1-name": "Flens",
                "weakpoint_set-2-name": "Kijkglas",
                "management-referer": base_url + '1/weakpoints/'
            }
            redirect_url = base_url + '1/weakpoints/'
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Weakpoint.objects.order_by('-id')[0].delete(testcase.user, override=True)
                super(section_weakpoints_edit, self).posttest(testcase, allowed)

        class section_subsections_edit(utils_tests.TestDefinition):
            # this is only the get test
            description = 'section tab subsections edit'
            url = base_url + '1/subsections/edit/'
            users = [1,]

        class section_weakpoints_delete(utils_tests.edit_TestDefinition):
            description = 'delete a weakpoint, test deleted flag'
            url = base_url + '1/weakpoints/edit/'
            users = [1,]
            data = {
                "weakpoint_set-TOTAL_FORMS": "5",
                "weakpoint_set-INITIAL_FORMS": "2",
                "weakpoint_set-MAX_NUM_FORMS": "0",
                "weakpoint_set-0-id": 1,
                "weakpoint_set-0-name": "Drainpunt",
                "weakpoint_set-1-id": 2,
                "weakpoint_set-1-name": "Flens",
                "weakpoint_set-1-DELETE": True,
                "management-referer": base_url + '1/weakpoints/'
            }
            redirect_url = base_url + '1/weakpoints/'
            def pretest(self, testcase, allowed):
                testcase.assertEquals(app_models.Weakpoint.objects.count(), 2)
                super(section_weakpoints_delete, self).pretest(testcase, allowed)
            def posttest(self, testcase, allowed):
                testcase.assertEquals(app_models.Weakpoint.objects.count(), 2)
                if allowed:
                    testcase.assertEquals(app_models.Weakpoint.objects.filter(deleted=True).count(), 1)
                    # special way to retrieve the object here, because deleted objects are normally excluded
                    item = app_models.Weakpoint.objects.get(v_id=2, version__isnull=True)
                    item._version = WorkVersion
                    item.deleted = False
                    item.save(testcase.user)
                else:
                    testcase.assertEquals(app_models.Weakpoint.objects.filter(deleted=True).count(), 0)
                super(section_weakpoints_delete, self).posttest(testcase, allowed)

        class section_delete(utils_tests.delete_TestDefinition):
            description = 'delete a section'
            users = [1,]
            redirect_url = overview_url
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Section(name='to_delete')
                newitem.save(defaultuser)
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.v_id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Section.objects.get_versionedobject(id=self.pk, version=WorkVersion).delete(defaultuser, override=True)

        tests = [
            installations(),
            overview(),
            overview_edit(),
            overview_report(),
            overview_compare(),
            overview_history(),
            overview_tree_json(),
            section1_section(),
            section2_section(),
            section_weakpoints(),
            section_subsections(),
            section_substances(),
            section_reactions(),
            section_scenarios(),
            section1_permissions(),
            section2_permissions(),
            section_compare(),
            section_report(),
            section_add(),
            subsection_add(),
            section_edit(),
            section_weakpoints_edit(),
            section_subsections_edit(),
            section_weakpoints_delete(),
            section_delete(),
        ]

        utils_tests.run_tests(self, users, tests)

    def test_model_report(self):
        user = User.objects.get(pk=1)
        section = app_models.Section.objects.get(pk=1)
        self.assertTrue('\n'.join(section.get_report_body(user, 'en')).find('Section'))
        self.assertTrue('\n'.join(section.get_report_body(user, 'nl')).find('Sectie'))

class InstallationsSectionPublishTestCase(InstallationsTestCase):
    """
    Use special fixture with versioned substances
    """
    fixtures = ['dummyusers', 'versioning', 'installations_publish']
    def test_overview_publish(self):

        base_url = '/installations/sections/'
        overview_url = '/installations/overview/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3)

        class overview_publish(utils_tests.add_TestDefinition):
            description = 'installations overview publish'
            url = '/installations/publish/'
            users = [1,]
            data = {
                "name": "public version",
                "management-referer": overview_url
            }
            redirect_url = overview_url


        tests = [
            overview_publish(),
        ]

        utils_tests.run_tests(self, users, tests)

class InstallationsSectionAutoscopeTestCase(InstallationsTestCase):
    def test_section_autoscope(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Section.objects.count(), 2)
        self.assertEquals(app_models.Scope.objects.count(), 3)
        self.assertEquals(app_models.Scope_Section.objects.count(), 2)
        section = app_models.Section(name = 'test'
        )
        section.save(user)
        v_id = section.v_id
        # We should have a section, an autoscope and a scope_section
        self.assertEquals(app_models.Section.objects.count(), 3)
        self.assertEquals(app_models.Scope.objects.count(), 4)
        self.assertEquals(app_models.Scope_Section.objects.count(), 3)

        # reload
        section = app_models.Section.objects.get_versionedobject(v_id=v_id, version=WorkVersion)
        section.delete(user)

        # not versioned so the new item should disappear
        self.assertEquals(app_models.Section.objects.count(), 2)
        self.assertEquals(app_models.Scope.objects.count(), 3)
        self.assertEquals(app_models.Scope_Section.objects.count(), 2)

def print_tree(items):
    res =[]
    prefix = ''
    for element,structure in tree_item_iterator(items):
        if structure['new_level']:
            prefix = '-%s' % prefix
        res.append(u'%s%s' % (prefix, element))
        x = len(structure['closed_levels'])
        if x:
            prefix = prefix[:-x]
    return u'|'.join(res)

# TODO: tree tests on section subsections tab edit

class InstallationsSectionTreeTest(InstallationsTestCase):
    def test_sectiontree_views(self):
        base_url = '/installations/overview/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3)

        class move(utils_tests.edit_TestDefinition):
            description = 'move section'
            url = base_url + 'edit/'
            users = [1,]
            data = {
                "actions": "move|2|1|after;",
                "hash": "",
                "management-referer": base_url
            }
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
                self.data['hash'] = trees_views.security_hash(treeobjects)

            def extra_tests(self, testcase, response, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                if allowed:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|-Methanol opslag en verlading')
                else:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
            def posttest(self, testcase, allowed):
                # restore situation
                if allowed:
                    app_models.SectionTreeElement.objects.get(v_id=2, version__isnull=True).move_to(
                        defaultuser,
                        app_models.SectionTreeElement.objects.get(v_id=1, version__isnull=True), 'last-child')

        class addsection(utils_tests.edit_TestDefinition):
            description = 'add section'
            url = base_url + 'edit/'
            users = [1,]
            data = {
                "actions": "add|1|section|New section;",
                "hash": "",
                "management-referer": base_url
            }
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
                self.data['hash'] = trees_views.security_hash(treeobjects)

            def extra_tests(self, testcase, response, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                if allowed:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading|--New section')
                else:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
            def posttest(self, testcase, allowed):
                # restore situation
                if allowed:
                    x = app_models.SectionTreeElement.objects.filter(version__isnull=True).order_by('-id')[0]
                    x.delete(defaultuser)

        class rename(utils_tests.edit_TestDefinition):
            description = 'rename section'
            url = base_url + 'edit/'
            users = [1,]
            data = {
                "actions": "rename|2|Y;",
                "hash": "",
                "management-referer": base_url
            }
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
                self.data['hash'] = trees_views.security_hash(treeobjects)

            def extra_tests(self, testcase, response, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                if allowed:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Y')
                else:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
            def posttest(self, testcase, allowed):
                # restore situation
                if allowed:
                    x = app_models.SectionTreeElement.objects.get(v_id=2, version__isnull=True).section
                    x.name = 'Methanol opslag en verlading'
                    x.save(defaultuser)

        tests = [
            move(),
            addsection(),
            rename(),
        ]

        utils_tests.run_tests(self, users, tests)

class InstallationsSectionTreeComplexTest(InstallationsTestCase):
    def test_sectiontree_views(self):
        base_url = '/installations/overview/'
        defaultuser = User.objects.get(pk=1)
        # only test for one user
        users = (1,)

        class complex(utils_tests.edit_TestDefinition):
            description = 'complex sectiontree editing'
            url = base_url + 'edit/'
            users = [1,]
            data = {
                "actions": "add|1|section|New section;rename|-1|X;move|2|-1|after;",
                "hash": "",
                "management-referer": base_url
            }
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
                self.data['hash'] = trees_views.security_hash(treeobjects)

            def extra_tests(self, testcase, response, allowed):
                treeobjects = list(app_models.SectionTreeElement.tree.all(
                    ).get_versionedobjects(version=WorkVersion).select_related_versioned())
                if allowed:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--X|--Methanol opslag en verlading')
                else:
                    testcase.assertEquals(print_tree(treeobjects), '-Formaldehyde installatie|--Methanol opslag en verlading')
            def posttest(self, testcase, allowed):
                # No restore
                pass


        tests = [
            complex(),
        ]

        utils_tests.run_tests(self, users, tests)

class InstallationsSectionSubstanceTestCase(InstallationsTestCase):
    def test_section_substance_views(self):
        base_url = '/installations/sections/substances/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3)

        class section_substance(utils_tests.TestDefinition):
            description = 'display section_substance'
            url = base_url + '1/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/edit/', 'users': [1,]},
                {'url': base_url + '1/delete/', 'users': [1,]},
            ]

        class section_substance_compare(utils_tests.TestDefinition):
            description = 'compare a section_substance'
            url = base_url + '1/compare/0/'
            users = [1,2,]

        class section_substance_add(utils_tests.add_TestDefinition):
            description = 'add a section_substance'
            url = '/installations/sections/1/substance/add/'
            users = [1,]
            data = {
                "substance": 2,
                "section": 1,
                "management-referer": '/installations/sections/1/substances/'
            }
            redirect_url = '/installations/sections/1/substances/'
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section_Substance.objects.order_by('-id')[0].delete(testcase.user, override=True)

        class section_substance_edit(utils_tests.edit_TestDefinition):
            description = 'edit a section_substance'
            url = base_url + '1/edit/'
            users = [1,]
            data = {
                "substance": 1,
                "section": 1,
                "management-referer": base_url + '1/'
            }
            redirect_url = base_url + '1/'

        class section_substance_delete(utils_tests.delete_TestDefinition):
            description = 'delete a section_substance'
            users = [1,]
            redirect_url = '/installations/sections/1/substances/'
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Section_Substance(section=1, substance=1)
                newitem.save(defaultuser)
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.v_id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Section_Substance.objects.get(id=self.pk).delete(defaultuser, override=True)


        tests = [
            section_substance(),
            section_substance_compare(),
            section_substance_add(),
            section_substance_edit(),
            section_substance_delete(),
        ]

        utils_tests.run_tests(self, users, tests)

class InstallationsSectionReactionTestCase(InstallationsTestCase):
    def test_section_reaction_views(self):
        base_url = '/installations/sections/reactions/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3)

        class section_reaction(utils_tests.TestDefinition):
            description = 'display section_reaction'
            url = base_url + '1/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/edit/', 'users': [1,]},
                {'url': base_url + '1/delete/', 'users': [1,]},
            ]

        class section_reaction_compare(utils_tests.TestDefinition):
            description = 'compare a section_reaction'
            url = base_url + '1/compare/0/'
            users = [1,2,]

        class section_reaction_add(utils_tests.add_TestDefinition):
            description = 'add a section_reaction'
            url = '/installations/sections/1/reaction/add/'
            users = [1,]
            data = {
                "reaction": 2,
                "section": 1,
                "management-referer": '/installations/sections/1/reactions/'
            }
            redirect_url = '/installations/sections/1/reactions/'
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section_Reaction.objects.order_by('-id')[0].delete(testcase.user, override=True)

        class section_reaction_edit(utils_tests.edit_TestDefinition):
            description = 'edit a section_reaction'
            url = base_url + '1/edit/'
            users = [1,]
            data = {
                "reaction": 1,
                "section": 1,
                "management-referer": base_url + '1/'
            }
            redirect_url = base_url + '1/'

        class section_reaction_delete(utils_tests.delete_TestDefinition):
            description = 'delete a section_reaction'
            users = [1,]
            redirect_url = '/installations/sections/1/reactions/'
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Section_Reaction(section=1, reaction=1)
                newitem.save(defaultuser)
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.v_id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Section_Reaction.objects.get(id=self.pk).delete(defaultuser, override=True)


        tests = [
            section_reaction(),
            section_reaction_compare(),
            section_reaction_add(),
            section_reaction_edit(),
            section_reaction_delete(),
        ]

        utils_tests.run_tests(self, users, tests)

class InstallationsScopeTestCase(InstallationsTestCase):
    def test_scope_views(self):

        base_url = '/installations/scopes/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3)

        class scopes(utils_tests.tablelist_TestDefinition):
            description = 'list of scopes'
            url = base_url
            users = [1,2,]
            contains = [{'url': base_url + 'add/', 'users': [1,]},]

        class scope1_scope(utils_tests.TestDefinition):
            description = 'display scope 1 tab scope'
            # no edit or delete because 1 is an autoscope
            url = base_url + '1/scope/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/scope/edit/', 'users': []},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class scope3_scope(utils_tests.TestDefinition):
            description = 'display scope 3  tab scope'
            url = base_url + '3/scope/'
            users = [1,2,]
            contains = [
                {'url': base_url + '3/scope/edit/', 'users': [1,]},
                {'url': base_url + '3/delete/', 'users': [1,]},
            ]

        class scope_sections(utils_tests.TestDefinition):
            description = 'display scope tab sections'
            # no edit or delete because 1 is an autoscope
            url = base_url + '1/sections/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/sections/edit/', 'users': []},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class scope_scenarios(utils_tests.TestDefinition):
            description = 'display scope tab scenarios'
            url = base_url + '1/scenarios/'
            users = [1,2,]
            contains = [
                {'url': base_url + '1/edit/', 'users': []},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class scope_compare(utils_tests.TestDefinition):
            description = 'compare a scope'
            url = base_url + '1/compare/0/'
            users = [1,2,]

        class scope_add(utils_tests.add_TestDefinition):
            description = 'add a scope'
            url = base_url + 'add/'
            users = [1,]
            data = {
                "name": "bogus",
                "description": "",
                "management-referer": base_url
            }
            def pretest(self, testcase, allowed):
                newvid = app_models.Scope.objects.next_v_id()
                self.redirect_url = base_url + '%s/sections/' % newvid
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Scope.objects.order_by('-id')[0].delete(testcase.user, override=True)

        class scope_scope_edit(utils_tests.edit_TestDefinition):
            description = 'edit a scope'
            url = base_url + '3/scope/edit/'
            users = [1,]
            data = {
                "name": "A1",
                "description": "",
                "management-referer": base_url + '3/scope/'
            }
            redirect_url = base_url + '3/scope/'

        class scope_sections_edit(utils_tests.edit_TestDefinition):
            description = 'edit sections for scope'
            users = [1,]
            data = {
                "actions": "setsection|1;",
                "hash": "",
                "management-referer": base_url
            }
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                from installations.views.main import get_visible_objects
                # create a new item to which we can add
                newitem = app_models.Scope(name='using this scope')
                newitem.save(defaultuser)
                self.v_id = newitem.v_id
                self.url = base_url + '%s/sections/edit/' % self.v_id
                testcase.assertEquals(app_models.Scope_Section.objects.filter(scope=self.v_id, deleted=False).count(), 0)
                testcase.assertEquals(app_models.Scope_Section.objects.filter(scope=self.v_id, deleted=True).count(), 0)

                if allowed:
                    queryset = app_models.SectionTreeElement.get_authorized_tree(testcase.user).get_versionedobjects(WorkVersion)
                    for item in queryset:
                        item.selected = False
                    res = get_visible_objects(queryset)
                    annotate_versioned(res, 'section')
##                    treeobjects = list(app_models.SectionTreeElement.get_authorized_tree(testcase.user).get_versionedobjects(version=WorkVersion).select_related_versioned())
                    self.data['hash'] = trees_views.security_hash(res)

                super(scope_sections_edit, self).pretest(testcase, allowed)

            def extra_tests(self, testcase, response, allowed):
                if allowed:
                    testcase.assertEquals(app_models.Scope_Section.objects.filter(scope=self.v_id, deleted=False).count(), 1)
                    testcase.assertEquals(app_models.Scope_Section.objects.filter(scope=self.v_id, deleted=True).count(), 0)
                else:
                    testcase.assertEquals(app_models.Scope_Section.objects.filter(scope=self.v_id, deleted=False).count(), 0)
                    testcase.assertEquals(app_models.Scope_Section.objects.filter(scope=self.v_id, deleted=True).count(), 0)
            def posttest(self, testcase, allowed):
                # restore situation
                app_models.Scope.objects.get_versionedobject(v_id=self.v_id, version=WorkVersion).delete(defaultuser, override=True)

        class scope_delete(utils_tests.delete_TestDefinition):
            description = 'delete a scope'
            users = [1,]
            redirect_url = base_url
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Scope(name='to_delete')
                newitem.save(defaultuser)
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.v_id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Scope.objects.get_versionedobject(id=self.pk, version=WorkVersion).delete(defaultuser, override=True)

        class scope_report(utils_tests.edit_TestDefinition):
            description = 'pdf report for scope'
            url = base_url + '1/pdf/'
            users = [1,2,]
            data = {
                "submit_pdf": "Generate pdf",
            }

        tests = [
            scopes(),
            scope1_scope(),
            scope3_scope(),
            scope_sections(),
            scope_scenarios(),
            scope_compare(),
            scope_add(),
            scope_scope_edit(),
            scope_sections_edit(),
            scope_delete(),
            scope_report(),
        ]


##        print [ (item.scope_id, item.section_id) for item in app_models.Scope_Section.objects.all()]
##        print [ (item.v_id, item.name) for item in app_models.Scope.objects.all()]
##        print [ (item.v_id, item.name) for item in app_models.Section.objects.all()]
        utils_tests.run_tests(self, users, tests)

class InstallationsVersioningTestCase(TestCase):
    fixtures = ['dummyusers','hazards_versioning', 'installations_versioning']

    def setUp(self):
        from django.contrib.auth.models import User, Permission
        from django.contrib.contenttypes.models import ContentType
        # give the users permissions

        # authorisations for users:
        # 1 = authorised
        # 2 = authorised to view all versions
        # 3 = authorised to view public versions
        # 4 = unauthorised

        u = User.objects.get(pk=1)
        permissions = [
            ('display_versions','versioning','version'),
            ('display_hazards','hazards','substance'),
            ('display_installations','installations','section'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

        u = User.objects.get(pk=2)
        permissions = [
            ('display_versions','versioning','version'),
            ('display_hazards','hazards','substance'),
            ('display_installations','installations','section'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

        u = User.objects.get(pk=3)
        permissions = [
            ('display_published_versions','versioning','version'),
            ('display_hazards','hazards','substance'),
            ('display_installations','installations','section'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

class InstallationsSectionVersioningTestCase(InstallationsVersioningTestCase):
    fixtures = ['dummyusers', 'installations_versioning']
    def test_section_versioning(self):
        user = User.objects.get(pk=1)

        # current state: five sections unversioned, all of them are dirty
        self.assertEquals(app_models.Section.objects.count(), 5)
        self.assertEquals(app_models.Section.objects.dirty().count(), 5)

        v1 = Version(name='v1')
        v1.save(user)

        # current state: five sections unversioned, five versioned
        self.assertEquals(app_models.Section.objects.count(), 10)
        self.assertEquals(app_models.Section.objects.dirty().count(), 0)
        A_u = app_models.Section.objects.get(v_id=1, version__isnull=True)
        A_v1 = app_models.Section.objects.get(v_id=1, version=v1)

        self.assertEquals(app_models.Section.objects.next_v_id(), 6)

        # let's add a section
        newsection = app_models.Section(name='F')
        newsection.save(user, parent = A_u.sectiontreeelement_set.all()[0])
        self.assertEquals(newsection.v_id, 6)

        self.assertEquals(app_models.Section.objects.count(), 11)
        self.assertEquals(app_models.Section.objects.dirty().count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.Section.objects.count(), 12)
        self.assertEquals(app_models.Section.objects.dirty().count(), 0)

        # let's change B
        s = app_models.Section.objects.get(name='B', version__isnull=True)
        s.description = 'newB'
        s.save(user)

        self.assertEquals(app_models.Section.objects.count(), 12)
        self.assertEquals(app_models.Section.objects.dirty().count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        self.assertEquals(app_models.Section.objects.count(), 13)
        self.assertEquals(app_models.Section.objects.dirty().count(), 0)

        # let's delete D
        s = app_models.Section.objects.get(name='D', version__isnull=True)
        s.delete(user)

        self.assertEquals(app_models.Section.objects.count(), 13)
        self.assertEquals(app_models.Section.objects.dirty().count(), 1)
        self.assertEquals(app_models.Section.objects.filter(deleted=True).count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v4 = Version(name='v4')
        v4.save(user)

        self.assertEquals(app_models.Section.objects.count(), 13)
        self.assertEquals(app_models.Section.objects.dirty().count(), 0)
        self.assertEquals(app_models.Section.objects.filter(deleted=True).count(), 1)

        # let's add a section and delete it again
        A_u = app_models.Section.objects.get(v_id=1, version__isnull=True)
        newsection = app_models.Section(name='G')
        newsection.save(user, parent = A_u.sectiontreeelement_set.all()[0])
        self.assertEquals(newsection.v_id, 7)
        self.assertEquals(app_models.Section.objects.count(), 14)
        self.assertEquals(app_models.Section.objects.dirty().count(), 1)
        s = app_models.Section.objects.get(name='G')
        s.delete(user)
        # G should dissappear entirely
        self.assertEquals(app_models.Section.objects.count(), 13)
        self.assertEquals(app_models.Section.objects.dirty().count(), 0)


        # now let's check the manager

        qs = app_models.Section.objects.all().get_versionedobjects(version=WorkVersion).order_by('name')
        # A, B, C, E and F
        self.assertEquals(qs.count(), 5)
        self.assertEquals(qs.__repr__(), '[<Section: A>, <Section: B>, <Section: C>, <Section: E>, <Section: F>]')
        self.assertEquals(qs[1].description, 'newB')

        qs = app_models.Section.objects.all().get_versionedobjects(version=v1).order_by('name')
        # original items
        self.assertEquals(qs.count(), 5)
        self.assertEquals(qs.__repr__(), '[<Section: A>, <Section: B>, <Section: C>, <Section: D>, <Section: E>]')
        self.assertEquals(qs[1].description, None)

        qs = app_models.Section.objects.all().get_versionedobjects(version=v2).order_by('name')
        # including section F
        self.assertEquals(qs.count(), 6)
        self.assertEquals(qs.__repr__(), '[<Section: A>, <Section: B>, <Section: C>, <Section: D>, <Section: E>, <Section: F>]')

        qs = app_models.Section.objects.all().get_versionedobjects(version=v3).order_by('name')
        # B changed
        self.assertEquals(qs.count(), 6)
        self.assertEquals(qs.__repr__(), '[<Section: A>, <Section: B>, <Section: C>, <Section: D>, <Section: E>, <Section: F>]')
        self.assertEquals(qs[1].description, 'newB')

        qs = app_models.Section.objects.all().get_versionedobjects(version=v4).order_by('name')
        # without D
        self.assertEquals(qs.count(), 5)
        self.assertEquals(qs.__repr__(), '[<Section: A>, <Section: B>, <Section: C>, <Section: E>, <Section: F>]')

        # check collect_items
        self.assertEquals(len(v1.collect_items(user)), 5) # all sections
        self.assertEquals(len(v2.collect_items(user)), 2) # section + tree (not autoscope + section_scope)
        self.assertEquals(len(v3.collect_items(user)), 1) # only section B
        self.assertEquals(len(v4.collect_items(user)), 2) # section + tree (no scopes in fixture --> no extra items here)

        # check moving children when deleting
        # add an element to C
        s = app_models.Section.objects.get_versionedobject(name='C', version=WorkVersion)
        sel = s.sectiontreeelement_set.all()[0]
        newsection = app_models.Section(name='G')
        newsection.save(user, parent = sel)
        self.assertEquals(len(sel.get_children()), 2)
        self.assertEquals(sel.get_children().__repr__(), '[<SectionTreeElement: E>, <SectionTreeElement: G>]')
        self.assertEquals(len(sel.parent.get_children()), 3)
        self.assertEquals(sel.parent.get_children().__repr__(), '[<SectionTreeElement: B>, <SectionTreeElement: C>, <SectionTreeElement: F>]')

        s.delete(user)

        a = app_models.Section.objects.get_versionedobject(name='A', version=WorkVersion)
        ael = a.sectiontreeelement_set.all()[0]
        self.assertEquals(len(ael.get_children()), 4)
        self.assertEquals(ael.get_children().__repr__(), '[<SectionTreeElement: B>, <SectionTreeElement: F>, <SectionTreeElement: E>, <SectionTreeElement: G>]')

class InstallationsPublishTestCase(InstallationsVersioningTestCase):
    def test_publish_installations(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Section.objects.count(), 5)
        self.assertEquals(app_models.Section.objects.dirty().count(), 5)
        self.assertEquals(hazards_models.Substance.objects.count(), 2)
        self.assertEquals(hazards_models.Substance.objects.dirty().count(), 2)

        # publish everything
        v1 = Version(name='v1', published=True)
        v1.save(user)

        self.assertEquals(app_models.Section.objects.count(), 10)
        self.assertEquals(app_models.Section.objects.dirty().count(), 0)
        self.assertEquals(hazards_models.Substance.objects.count(), 4)
        self.assertEquals(hazards_models.Substance.objects.dirty().count(), 0)

        # change section 1
        s = app_models.Section.objects.get(v_id=1, version__isnull=True)
        self.assertEquals(s.name, 'A')
        s.name= 'Aa'
        s.save(user)

        # change substance 1
        su = hazards_models.Substance.objects.get(v_id=1, version__isnull=True)
        self.assertEquals(su.name, 'A')
        su.name= 'Aa'
        su.save(user)

        self.assertEquals(app_models.Section.objects.dirty().count(), 1)
        self.assertEquals(app_models.Section.objects.filter(published_dirty=True, version__isnull=True).count(), 1)

        self.assertEquals(hazards_models.Substance.objects.dirty().count(), 1)
        self.assertEquals(hazards_models.Substance.objects.filter(published_dirty=True, version__isnull=True).count(), 1)

        # version installations info
        v2 = Version(name='v2', published=True)
        v2.save(user, propagate=False)
        app_models.publish_installations(user, v2)

        self.assertEquals(app_models.Section.objects.dirty().count(), 0)
        self.assertEquals(app_models.Section.objects.filter(published_dirty=True, version__isnull=True).count(), 0)

        self.assertEquals(hazards_models.Substance.objects.dirty().count(), 1)
        self.assertEquals(hazards_models.Substance.objects.filter(published_dirty=True, version__isnull=True).count(), 1)

        # let's check the substance in the published v2
        su1 = hazards_models.Substance.objects.get_versionedobject(v_id=1, version=v2)
        self.assertEquals(su1.name, 'A')


class InstallationsSectionSubstanceVersioningTestCase(InstallationsVersioningTestCase):
    def test_section_substance_versioning(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Section_Substance.objects.count(), 0)
        self.assertEquals(app_models.Section.objects.count(), 5)
        self.assertEquals(hazards_models.Substance.objects.count(), 2)

        # creating a new section_substance
        s1 = app_models.Section.objects.get(pk=1)
        sub1 = hazards_models.Substance.objects.get(pk=1)
        newitem = app_models.Section_Substance(section=s1, substance=sub1)
        newitem.save(user)

        self.assertEquals(app_models.Section_Substance.objects.count(), 1)
        self.assertEquals(app_models.Section_Substance.objects.dirty().count(), 1)

        # version everything
        v1 = Version(name='v1')
        v1.save(user)

        self.assertEquals(app_models.Section_Substance.objects.count(), 2)
        self.assertEquals(app_models.Section_Substance.objects.dirty().count(), 0)

        self.assertEquals(app_models.Section.objects.count(), 10)
        self.assertEquals(hazards_models.Substance.objects.count(), 4)

        # the versioned section_substance refers to the versioned section and the versioned substance
        A_u = app_models.Section_Substance.objects.get(v_id=1, version__isnull=True)
        A_v1 = app_models.Section_Substance.objects.get(v_id=1, version=v1)
        self.assertNotEquals(A_u.section.pk, A_v1.section.pk)
        self.assertNotEquals(A_u.substance.pk, A_v1.substance.pk)

        # make the section_substance dirty
        newitem.aggregation = 'liquid'
        newitem.save(user)

        # create a new version
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.Section_Substance.objects.count(), 3)
        self.assertEquals(app_models.Section_Substance.objects.dirty().count(), 0)

        # substance and section not in new version
        self.assertEquals(app_models.Section.objects.count(), 10)
        self.assertEquals(hazards_models.Substance.objects.count(), 4)

        A_u = app_models.Section_Substance.objects.get(v_id=1, version__isnull=True)
        A_v1 = app_models.Section_Substance.objects.get(v_id=1, version=v1)
        A_v2 = app_models.Section_Substance.objects.get(v_id=1, version=v2)

        self.assertNotEquals(A_u.section.pk, A_v2.section.pk)
        self.assertNotEquals(A_u.substance.pk, A_v2.substance.pk)
        # the new versioned item refers to the same objects as the previous version since they were not reversioned
        self.assertEquals(A_v1.section.pk, A_v2.section.pk)
        self.assertEquals(A_v1.substance.pk, A_v2.substance.pk)

        # Reverse

        # make the substance dirty
        sub1 = hazards_models.Substance.objects.get(v_id=1, version__isnull=True)
        sub1.name = 'A2'
        sub1.save(user)

        # create a new version
        v3 = Version(name='v3')
        v3.save(user)

        self.assertEquals(hazards_models.Substance.objects.count(), 5)
        self.assertEquals(hazards_models.Substance.objects.dirty().count(), 0)

        # substance and section_substance not in new version
        self.assertEquals(app_models.Section.objects.count(), 10)
        self.assertEquals(app_models.Section_Substance.objects.count(), 3)

        sub1 = hazards_models.Substance.objects.get_versionedobject(v_id=1, version=v1)
        sub2 = hazards_models.Substance.objects.get_versionedobject(v_id=1, version=v2)
        sub3 = hazards_models.Substance.objects.get_versionedobject(v_id=1, version=v3)

        # the substance in v1 and v2 are the same, in v3 is different
        self.assertEquals(sub1.pk, sub2.pk)
        self.assertNotEquals(sub1.pk, sub3.pk)
        self.assertNotEquals(sub2.pk, sub3.pk)
        # sub3 and sub2 refer to the same objects since they were not reversioned
        self.assertEquals(sub3.section_substance_set.all()[0].pk, sub2.section_substance_set.all()[0].pk)
        # but sub1 refers to another object
        self.assertNotEquals(sub3.section_substance_set.all()[0].pk, sub1.section_substance_set.all()[0].pk)




class InstallationsPermissionsTestCase(TestCase):
    fixtures = ['dummyusers', 'hazards', 'installations_permissions']

    def setUp(self):
        from django.contrib.auth.models import User, Permission
        from django.contrib.contenttypes.models import ContentType
        # give the users permissions

        # authorisations for users:
        # 1 = authorised
        # 2 = unauthorised to edit, but authorized on the second installation
        # 3 = unauthorised to edit
        # 4 = unauthorised to even display

        u = User.objects.get(pk=1)

        permissions = [
            ('display_versions','versioning','version'),
            ('display_installations','installations','section'),
            ('edit_installations','installations','section'),
            ('display_deviationscenarios','deviations','deviationscenario'),
            ('edit_deviationscenarios','deviations','deviationscenario'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

        u = User.objects.get(pk=2)
        permissions = [
            ('display_versions','versioning','version'),
            ('display_installations','installations','section'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

        u = User.objects.get(pk=3)
        permissions = [
            ('display_versions','versioning','version'),
            ('display_installations','installations','section'),
        ]
        for item in permissions:
            u.user_permissions.add(Permission.objects.get_by_natural_key(*item))

class InstallationsPermissionsSectionTestCase(InstallationsPermissionsTestCase):
    def test_section_views(self):

        base_url = '/installations/sections/'
        overview_url = '/installations/overview/'
        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3,4)

        class overview(utils_tests.TestDefinition):
            description = 'installations overview'
            url = overview_url
            users = [1,2,3,]
            contains = [{'url': overview_url + 'edit/', 'users': [1,2,]},
            # add no longer on overview
                {'url': base_url + 'add/', 'users': []},]

        class overview_edit(utils_tests.TestDefinition):
            # this is only the get test
            description = 'installations overview edit'
            url = overview_url + 'edit/'
            users = [1,2, ]
            def extra_tests(self, testcase, response, allowed):
                if allowed:
                    if testcase.user.id == 1:
                        testcase.assertEquals(len(response.context['treeobjects']), 4)
                    elif testcase.user.id == 2:
                        testcase.assertEquals(len(response.context['treeobjects']), 2)

        class section2_section(utils_tests.TestDefinition):
            description = 'display section 2 tab section'
            url = base_url + '2/section/'
            users = [1,2,3]
            contains = [
                {'url': base_url + '2/section/edit/', 'users': [1,]},
                {'url': base_url + '2/delete/', 'users': [1,]},
            ]

        class section4_section(utils_tests.TestDefinition):
            description = 'display section 4 tab section'
            url = base_url + '4/section/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '4/section/edit/', 'users': [1,2]},
                {'url': base_url + '4/delete/', 'users': [1,2]},
            ]

        class section2_weakpoints(utils_tests.TestDefinition):
            description = 'display section 2 tab weakpoints'
            url = base_url + '2/weakpoints/'
            users = [1,2,3]
            contains = [
                {'url': base_url + '2/weakpoints/edit/', 'users': [1,]},
                {'url': base_url + '2/delete/', 'users': []},
            ]

        class section4_weakpoints(utils_tests.TestDefinition):
            description = 'display section 4 tab weakpoints'
            url = base_url + '4/weakpoints/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '4/weakpoints/edit/', 'users': [1,2]},
                {'url': base_url + '4/delete/', 'users': []},
            ]

        class section2_substances(utils_tests.TestDefinition):
            description = 'display section 2 tab substances'
            url = base_url + '2/substances/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '2/substances/edit/', 'users': []},
                {'url': base_url + '2/delete/', 'users': []},
                {'url': base_url + '2/substance/add/', 'users': [1,]},
            ]

        class section4_substances(utils_tests.TestDefinition):
            description = 'display section 4 tab substances'
            url = base_url + '4/substances/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '4/substances/edit/', 'users': []},
                {'url': base_url + '4/delete/', 'users': []},
                {'url': base_url + '4/substance/add/', 'users': [1,2,]},
            ]

        class section2_reactions(utils_tests.TestDefinition):
            description = 'display section 2 tab reactions'
            url = base_url + '2/reactions/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '2/reactions/edit/', 'users': []},
                {'url': base_url + '2/delete/', 'users': []},
                {'url': base_url + '2/reaction/add/', 'users': [1,]},
            ]

        class section4_reactions(utils_tests.TestDefinition):
            description = 'display section 4 tab reactions'
            url = base_url + '4/reactions/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '4/reactions/edit/', 'users': []},
                {'url': base_url + '4/delete/', 'users': []},
                {'url': base_url + '4/reaction/add/', 'users': [1,2,]},
            ]

        class section2_scenarios(utils_tests.TestDefinition):
            description = 'display section 2 tab scenarios'
            url = base_url + '2/scenarios/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '2/scenarios/edit/', 'users': []},
                {'url': base_url + '2/delete/', 'users': []},
                {'url': base_url + '2/scenario/add/', 'users': [1,]},
            ]

        class section4_scenarios(utils_tests.TestDefinition):
            # add button depends on permissions on scenarios, not on section permission
            description = 'display section 4 tab scenarios'
            url = base_url + '4/scenarios/'
            users = [1,2,3,]
            contains = [
                {'url': base_url + '4/scenarios/edit/', 'users': []},
                {'url': base_url + '4/delete/', 'users': []},
                {'url': base_url + '4/scenario/add/', 'users': [1,]},
            ]

        class section1_permissions(utils_tests.TestDefinition):
            description = 'display section 1 tab permissions'
            url = base_url + '1/permissions/'
            users = [1,]
            contains = [
                {'url': base_url + '1/permissions/edit/', 'users': [1,]},
                {'url': base_url + '1/delete/', 'users': []},
            ]

        class section2_permissions(utils_tests.TestDefinition):
            description = 'display section 2 tab permissions'
            url = base_url + '2/permissions/'
            # This should not be available to anyone
            users = []
            contains = [
                {'url': base_url + '2/permissions/edit/', 'users': []},
                {'url': base_url + '2/delete/', 'users': []},
            ]

        class section3_permissions(utils_tests.TestDefinition):
            description = 'display section 1 tab permissions'
            url = base_url + '3/permissions/'
            users = [1,2,]
            contains = [
                {'url': base_url + '3/permissions/edit/', 'users': [1,2,]},
                {'url': base_url + '3/delete/', 'users': []},
            ]

        class section4_permissions(utils_tests.TestDefinition):
            description = 'display section 4 tab permissions'
            url = base_url + '4/permissions/'
            # This should not be available to anyone
            users = []
            contains = [
                {'url': base_url + '4/permissions/edit/', 'users': []},
                {'url': base_url + '4/delete/', 'users': []},
            ]

        class section_add(utils_tests.add_TestDefinition):
            description = 'add a section'
            url = base_url + 'add/'
            users = [1,]
            data = {
                "name": "bogus",
                "description": "",
                "management-referer": overview_url
            }
            redirect_url = overview_url
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section.objects.order_by('-id').get_versionedobjects(version=WorkVersion)[0].delete(testcase.user, override=True)

        class subsection1_add(utils_tests.add_TestDefinition):
            description = 'add a subsection to section 1'
            url = base_url + '1/section/add/'
            users = [1,]
            data = {
                "name": "bogus",
                "description": "",
                "management-referer": base_url + '1/section/'
            }
            redirect_url = base_url + '1/section/'
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section.objects.order_by('-id')[0].delete(testcase.user, override=True)

        class subsection3_add(utils_tests.add_TestDefinition):
            description = 'add a subsection to section 3'
            url = base_url + '3/section/add/'
            users = [1,2,]
            data = {
                "name": "bogus",
                "description": "",
                "management-referer": base_url + '3/section/'
            }
            redirect_url = base_url + '3/section/'
            def posttest(self, testcase, allowed):
                if allowed:
                    app_models.Section.objects.order_by('-id')[0].delete(testcase.user, override=True)

        class section2_edit(utils_tests.edit_TestDefinition):
            description = 'edit section 2'
            url = base_url + '2/section/edit/'
            users = [1,]
            data = {
                "name": "A1",
                "description": "",
                "management-referer": base_url + '2/section/'
            }
            redirect_url = base_url + '2/section/'

        class section4_edit(utils_tests.edit_TestDefinition):
            description = 'edit section 4'
            url = base_url + '4/section/edit/'
            users = [1,2,]
            data = {
                "name": "A1",
                "description": "",
                "management-referer": base_url + '4/section/'
            }
            redirect_url = base_url + '4/section/'

        class section2_weakpoints_edit(utils_tests.edit_TestDefinition):
            description = 'edit the weakpoints for section 2'
            url = base_url + '2/weakpoints/edit/'
            users = [1,]
            data = {
                "weakpoint_set-TOTAL_FORMS": "3",
                "weakpoint_set-INITIAL_FORMS": "0",
                "weakpoint_set-MAX_NUM_FORMS": "0",
                "management-referer": base_url + '2/weakpoints/'
            }
            redirect_url = base_url + '2/weakpoints/'

        class section4_weakpoints_edit(utils_tests.edit_TestDefinition):
            description = 'edit the weakpoints for section 4'
            url = base_url + '4/weakpoints/edit/'
            users = [1,2,]
            data = {
                "weakpoint_set-TOTAL_FORMS": "3",
                "weakpoint_set-INITIAL_FORMS": "0",
                "weakpoint_set-MAX_NUM_FORMS": "0",
                "management-referer": base_url + '4/weakpoints/'
            }
            redirect_url = base_url + '4/weakpoints/'

        class section1_permissions_edit(utils_tests.edit_TestDefinition):
            description = 'edit section 1 tab permissions'
            url = base_url + '1/permissions/edit/'
            users = [1,]
            data = {
                "authorized_users": [],
                "management-referer": base_url + '1/permissions/'
            }
            redirect_url = base_url + '1/permissions/'

        class section2_permissions_edit(utils_tests.edit_TestDefinition):
            description = 'edit section 2 tab permissions'
            url = base_url + '2/permissions/edit/'
            users = []
            data = {
                "authorized_users": [],
                "management-referer": base_url + '2/permissions/'
            }
            redirect_url = base_url + '2/permissions/'

        class section3_permissions_edit(utils_tests.edit_TestDefinition):
            description = 'edit section 3 tab permissions'
            url = base_url + '3/permissions/edit/'
            users = [1,2,]
            data = {
                "authorized_users": [2,],
                "management-referer": base_url + '3/permissions/'
            }
            redirect_url = base_url + '3/permissions/'

        class section4_permissions_edit(utils_tests.edit_TestDefinition):
            description = 'edit section 4 tab permissions'
            url = base_url + '4/permissions/edit/'
            users = []
            data = {
                "authorized_users": [],
                "management-referer": base_url + '4/permissions/'
            }
            redirect_url = base_url + '4/permissions/'

        class section1_delete(utils_tests.delete_TestDefinition):
            description = 'delete a section'
            users = [1,]
            redirect_url = overview_url
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Section(name='to_delete')
                newitem.save(defaultuser, parent=app_models.SectionTreeElement.objects.get_versionedobject(v_id=1, version=WorkVersion))
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.v_id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Section.objects.get_versionedobject(id=self.pk, version=WorkVersion).delete(defaultuser, override=True)

        class section3_delete(utils_tests.delete_TestDefinition):
            description = 'delete a section'
            users = [1,2,]
            redirect_url = overview_url
            def pretest(self, testcase, allowed):
                # create a new item that we can delete
                newitem = app_models.Section(name='to_delete')
                newitem.save(defaultuser, parent=app_models.SectionTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion))
                self.pk = newitem.id
                self.url = base_url + '%s/delete/' % newitem.v_id
            def posttest(self, testcase, allowed):
                # clean up if the item was not deleted
                if not allowed:
                    app_models.Section.objects.get_versionedobject(id=self.pk, version=WorkVersion).delete(defaultuser, override=True)

        tests = [
            overview(),
            overview_edit(),
            section2_section(),
            section4_section(),
            section2_weakpoints(),
            section4_weakpoints(),
            section2_substances(),
            section4_substances(),
            section2_reactions(),
            section4_reactions(),
            section2_scenarios(),
            section4_scenarios(),
            section1_permissions(),
            section2_permissions(),
            section3_permissions(),
            section4_permissions(),
            section_add(),
            subsection1_add(),
            subsection3_add(),
            section2_edit(),
            section4_edit(),
            section2_weakpoints_edit(),
            section4_weakpoints_edit(),
            section1_permissions_edit(),
            section2_permissions_edit(),
            section3_permissions_edit(),
            section4_permissions_edit(),
            section1_delete(),
            section3_delete(),
        ]

        utils_tests.run_tests(self, users, tests)

class InstallationsSearchTestCase(InstallationsTestCase):
    def test_searches(self):
        users = (0,1,2,3,4)

        class base_search(utils_tests.TestDefinition):
            description = ''
            searchclass = None
            searchitems = []
            users = [1,2,]
            def runtests(self, testcase, allowed, contains, notcontains):
                user = testcase.user or AnonymousUser()
                for item in self.searchitems:
                    res = allowed and item['result'] or 0
                    testcase.assertEquals(len(self.searchclass().search(item['query'], user, WorkVersion, None)), res)

        class search_section(base_search):
            description = 'search section'
            searchclass = app_models.Installations_SectionSearch
            searchitems = [
                { "query": 'Formaldehyde', "result": 1},
                { "query": 'Quench', "result": 0},
            ]

        class search_weakpoint(base_search):
            description = 'search weak point'
            searchclass = app_models.Installations_WeakpointSearch
            searchitems = [
                { "query": 'Flens', "result": 1},
                { "query": 'Window', "result": 0},
            ]

        class search_scope(base_search):
            description = 'search scope'
            searchclass = app_models.Installations_ScopeSearch
            searchitems = [
                { "query": 'Formaldehyde', "result": 1},
                { "query": 'Quench', "result": 0},
            ]


        tests = [
            search_section(self),
            search_weakpoint(self),
            search_scope(self),
        ]

        utils_tests.run_tests(self, users, tests)

