"""
Links app

! Depends on the versioning app

Objects for now need to be VersionedModel subclasses

"""
from django.db import models
from django.db.models import signals as django_signals
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from utils.versioning import models as versioning_models

class LinkManager(versioning_models.VersionedModelManager):
    def links_for_object(self, obj):
        object_type = ContentType.objects.get_for_model(obj)
        version = getattr(obj, '_version', obj.version) or versioning_models.WorkVersion
        return self.filter(content_type__pk=object_type.id,
                           object_id=obj.v_id).get_versionedobjects(version=version)

class Link(versioning_models.VersionedModel):
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = versioning_models.VersionedGenericForeignKey('content_type', 'object_id')
    name = models.CharField(verbose_name = _('Name'), max_length=100)
    url = models.CharField(verbose_name =_('Url'), max_length=200)

    objects = LinkManager()

    class Meta:
        verbose_name = _('Link')
        verbose_name_plural = _('Links')
        ordering = ['-date_created']

    def __unicode__(self):
        return '%s' % self.name


    # urls

    def get_absolute_url(self):
        return self.content_object.get_absolute_url()

    @models.permalink
    def get_delete_url(self):
        return ('links_delete', (self.v_id,))


    # permissions

    @classmethod
    def may_add(cls, user, parent=None):
        return parent.may_change(user)

    def may_change(self, user):
        if self.version:
            return False
        return self.content_object.may_change(user)

    def may_delete(self, user):
        return self.may_change(user)



def make_links_version(sender, instance, objects, **kwargs):
    if instance.published:
        filterkwarg={'published_dirty': True, 'version__isnull': True}
    else:
        filterkwarg={'dirty': True, 'version__isnull': True}
    if objects:
        for typelist in objects:
            if typelist:
                object_type = ContentType.objects.get_for_model(typelist[0])
                for item in Link.objects.filter(**filterkwarg).filter(content_type__pk=object_type.id,
                           object_id__in=[item.v_id for item in typelist]):
                    item.make_version(instance)
    else:
        for item in Link.objects.filter(**filterkwarg):
            item.make_version(instance)

versioning_models.post_version.connect(make_links_version, sender=versioning_models.Version)
